"""app URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/4.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.contrib.auth.views import LogoutView
from django.urls import include, path
from two_factor.urls import urlpatterns as tf_urls

from core.views import (
    ads,
    aggregators,
    banks,
    base,
    linkgroups,
    notifications,
    payouts,
    platforms,
    profit,
    wallets,
)

# from core.views.stripe_callbacks import Callback

urlpatterns = [
    path("__debug__/", include("debug_toolbar.urls")),
    path("", base.Home.as_view(), name="home"),
    path("sapp/", admin.site.urls),
    # 2FA login urls
    path("", include(tf_urls)),
    path("accounts/signup/", base.Signup.as_view(), name="signup"),
    path("accounts/logout/", LogoutView.as_view(), name="logout"),
    # Notifications
    path(
        "notifications/<str:type>/update/",
        notifications.NotificationsUpdate.as_view(),
        name="notifications_update",
    ),
    # Aggregators
    path(
        "aggs/<str:type>/",
        aggregators.AggregatorList.as_view(),
        name="aggregators",
    ),
    path(
        "aggs/<str:type>/create/",
        aggregators.AggregatorCreate.as_view(),
        name="aggregator_create",
    ),
    path(
        "aggs/<str:type>/update/<str:pk>/",
        aggregators.AggregatorUpdate.as_view(),
        name="aggregator_update",
    ),
    path(
        "aggs/<str:type>/delete/<str:pk>/",
        aggregators.AggregatorDelete.as_view(),
        name="aggregator_delete",
    ),
    # Aggregator Requisitions
    path(
        "aggs/<str:type>/info/<str:pk>/",
        aggregators.ReqsList.as_view(),
        name="reqs",
    ),
    # Aggregator Account link flow
    path(
        "aggs/<str:type>/countries/<str:pk>/",
        aggregators.AggregatorCountriesList.as_view(),
        name="aggregator_countries",
    ),
    path(
        "aggs/<str:type>/countries/<str:pk>/<str:country>/banks/",
        aggregators.AggregatorCountryBanksList.as_view(),
        name="aggregator_country_banks",
    ),
    path(
        "aggs/<str:type>/link/<str:pk>/<str:bank>/",
        aggregators.AggregatorLinkBank.as_view(),
        name="aggregator_link",
    ),
    # Delete requisition
    path(
        "aggs/<str:type>/delete/<str:pk>/<str:req_id>/",
        aggregators.ReqDelete.as_view(),
        name="req_delete",
    ),
    # Requisition info
    path(
        "aggs/<str:type>/info/<str:pk>/<str:req_id>/",
        aggregators.ReqInfo.as_view(),
        name="req_info",
    ),
    # Request bank fetch
    path(
        "ops/bank_fetch/<str:pk>/",
        aggregators.RequestBankFetch.as_view(),
        name="bank_fetch",
    ),
    path(
        "ops/bank_fetch/",
        aggregators.RequestBankFetch.as_view(),
        name="bank_fetch",
    ),
    # Bank details by currency
    path(
        "banks/<str:type>/details/",
        banks.BanksCurrencies.as_view(),
        name="currencies",
    ),
    path(
        "banks/<str:type>/req/<str:aggregator_id>/<str:req_id>/",
        banks.BanksRequisitionUpdate.as_view(),
        name="requisition_update",
    ),
    path(
        "banks/<str:type>/req_delete/<str:pk>/",
        banks.BanksRequisitionDelete.as_view(),
        name="requisition_delete",
    ),
    # Bank balances
    path(
        "banks/<str:type>/balances/",
        banks.BanksBalances.as_view(),
        name="balances",
    ),
    # Transactions
    path(
        "banks/<str:type>/transactions/<str:aggregator_id>/<str:account_id>/",
        banks.BanksTransactions.as_view(),
        name="transactions",
    ),
    # Platforms
    path(
        "platforms/<str:type>/",
        platforms.PlatformList.as_view(),
        name="platforms",
    ),
    path(
        "platforms/<str:type>/create/",
        platforms.PlatformCreate.as_view(),
        name="platform_create",
    ),
    path(
        "platforms/<str:type>/update/<str:pk>/",
        platforms.PlatformUpdate.as_view(),
        name="platform_update",
    ),
    path(
        "platforms/<str:type>/delete/<str:pk>/",
        platforms.PlatformDelete.as_view(),
        name="platform_delete",
    ),
    # Trades
    path(
        "trades/<str:type>/",
        platforms.PlatformTrades.as_view(),
        name="trades",
    ),
    # Ads
    path(
        "ads/<str:type>/",
        ads.AdList.as_view(),
        name="ads",
    ),
    path(
        "ads/<str:type>/create/",
        ads.AdCreate.as_view(),
        name="ad_create",
    ),
    path(
        "ads/<str:type>/update/<str:pk>/",
        ads.AdUpdate.as_view(),
        name="ad_update",
    ),
    path(
        "ads/<str:type>/delete/<str:pk>/",
        ads.AdDelete.as_view(),
        name="ad_delete",
    ),
    path(
        "ops/ads/dist/",
        ads.AdDist.as_view(),
        name="ad_dist",
    ),
    path(
        "ops/ads/nuke/",
        ads.AdNuke.as_view(),
        name="ad_nuke",
    ),
    path(
        "ops/ads/redist/",
        ads.AdRedist.as_view(),
        name="ad_redist",
    ),
    path(
        "ops/ads/dedup/",
        ads.AdDedup.as_view(),
        name="ad_dedup",
    ),
    path(
        "ops/ads/cheat/",
        ads.Cheat.as_view(),
        name="cheat",
    ),
    path(
        "profit/<str:type>/",
        profit.Profit.as_view(),
        name="profit",
    ),
    # Wallets
    path(
        "wallets/<str:type>/",
        wallets.WalletList.as_view(),
        name="wallets",
    ),
    path(
        "wallets/<str:type>/create/",
        wallets.WalletCreate.as_view(),
        name="wallet_create",
    ),
    path(
        "wallets/<str:type>/update/<str:pk>/",
        wallets.WalletUpdate.as_view(),
        name="wallet_update",
    ),
    path(
        "operator_wallets/<str:type>/update/",
        wallets.OperatorWalletsUpdate.as_view(),
        name="operator_wallets_update",
    ),
    path(
        "wallets/<str:type>/delete/<str:pk>/",
        wallets.WalletDelete.as_view(),
        name="wallet_delete",
    ),
    # Payouts
    path(
        "payouts/<str:type>/",
        payouts.PayoutList.as_view(),
        name="payouts",
    ),
    path(
        "payouts/<str:type>/create/",
        payouts.PayoutCreate.as_view(),
        name="payout_create",
    ),
    path(
        "payouts/<str:type>/update/<str:pk>/",
        payouts.PayoutUpdate.as_view(),
        name="payout_update",
    ),
    path(
        "payouts/<str:type>/delete/<str:pk>/",
        payouts.PayoutDelete.as_view(),
        name="payout_delete",
    ),
    path(
        "payouts/action/delete_all/",
        payouts.PayoutDeleteAll.as_view(),
        name="payout_delete_all",
    ),
    # Link groups
    path(
        "links/<str:type>/",
        linkgroups.LinkGroupList.as_view(),
        name="linkgroups",
    ),
    path(
        "links/<str:type>/create/",
        linkgroups.LinkGroupCreate.as_view(),
        name="linkgroup_create",
    ),
    path(
        "links/<str:type>/update/<str:pk>/",
        linkgroups.LinkGroupUpdate.as_view(),
        name="linkgroup_update",
    ),
    path(
        "links/<str:type>/delete/<str:pk>/",
        linkgroups.LinkGroupDelete.as_view(),
        name="linkgroup_delete",
    ),
    path(
        "links/<str:type>/info/<str:pk>/",
        linkgroups.LinkGroupInfo.as_view(),
        name="linkgroup_info",
    ),
    path(
        "links/<str:type>/simulate/<str:pk>/",
        linkgroups.LinkGroupSimulation.as_view(),
        name="linkgroup_simulate",
    ),
    path(
        "links/<str:type>/withdraw/<str:pk>/",
        linkgroups.LinkGroupWithdraw.as_view(),
        name="linkgroup_withdraw",
    ),
] + static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)
